# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import os
import pathlib
import re
import socket

import pytest

import spnego
import spnego._ntlm as ntlm
import spnego.channel_bindings
import spnego.iov
from spnego._credential import CredentialCache
from spnego._ntlm_raw.crypto import md4
from spnego._ntlm_raw.messages import (
    Authenticate,
    AvId,
    Challenge,
    FileTime,
    Negotiate,
    NegotiateFlags,
    TargetInfo,
    Version,
)
from spnego._text import to_bytes, to_text
from spnego.exceptions import (
    BadBindingsError,
    BadMICError,
    InvalidTokenError,
    NoContextError,
    OperationNotAvailableError,
    SpnegoError,
    UnsupportedQop,
)

from .conftest import get_data


def test_get_credential_file_no_env_var():
    actual = ntlm._get_credential_file()
    assert actual is None


def test_get_credential_file_env_var_missing_file(tmpdir, monkeypatch):
    tmp_creds = os.path.join(to_text(tmpdir), "pÿspᴞӛgӫ TÈ$" ".creds")

    monkeypatch.setenv("NTLM_USER_FILE", tmp_creds)

    actual = ntlm._get_credential_file()
    assert actual is None


def test_get_credential_file(tmpdir, monkeypatch):
    tmp_creds = os.path.join(to_text(tmpdir), "pÿspᴞӛgӫ TÈ$" ".creds")
    with open(tmp_creds, mode="wb") as fd:
        fd.write(b"data")

    monkeypatch.setenv("NTLM_USER_FILE", tmp_creds)

    actual = ntlm._get_credential_file()
    assert actual == to_text(tmp_creds)


@pytest.mark.parametrize(
    "line, username, domain, lm_hash, nt_hash, input",
    [
        (
            "domain:Username:password",
            "username",
            "domain",
            "E52CAC67419A9A224A3B108F3FA6CB6D",
            "8846F7EAEE8FB117AD06BDD830B7586C",
            CredentialCache(username="domain\\username"),
        ),
        (
            "domain:Username:password\ndomain:other:pass2",
            "Username",
            "domain",
            "E52CAC67419A9A224A3B108F3FA6CB6D",
            "8846F7EAEE8FB117AD06BDD830B7586C",
            None,
        ),
        (
            "fake\ndomain:username:password",
            "username",
            "domain",  # newline or garbage data  in file won't fail
            "E52CAC67419A9A224A3B108F3FA6CB6D",
            "8846F7EAEE8FB117AD06BDD830B7586C",
            CredentialCache("domain\\username"),
        ),
        (
            ":username@DOMAIN.COM:password",
            "username@DOMAIN.COM",
            None,
            "E52CAC67419A9A224A3B108F3FA6CB6D",
            "8846F7EAEE8FB117AD06BDD830B7586C",
            CredentialCache("username@DOMAIN.COM"),
        ),
        (
            "testuser:1000:278623D830DABE161104594F8C2EF12B:C3C6F4FD8A02A6C1268F1A8074B6E7E0:[U]:LCT-1589398321",
            "testuser",
            None,
            "278623D830DABE161104594F8C2EF12B",
            "C3C6F4FD8A02A6C1268F1A8074B6E7E0",
            CredentialCache("testuser"),
        ),
        (
            "TESTDOM\\testuser:1000:4588C64B89437893AAD3B435B51404EE:65202355FA01AEF26B89B19E00F52679:[U]:LCT-1589398321",
            "testuser",
            "testdom",
            "4588C64B89437893AAD3B435B51404EE",
            "65202355FA01AEF26B89B19E00F52679",
            CredentialCache("testdom\\testuser"),
        ),
        (
            "TESTDOM\\testuser:1000:4588C64B89437893AAD3B435B51404EE:65202355FA01AEF26B89B19E00F52679:[U]:LCT-1589398321",
            "testuser",
            "testdom",
            "4588C64B89437893AAD3B435B51404EE",
            "65202355FA01AEF26B89B19E00F52679",
            CredentialCache("testdom\\testuser"),
        ),
        (
            "testuser@TESTDOM.COM:1000:00000000000000000000000000000000:8ADB9B997580D69E69CAA2BBB68F4697:[U]:LCT-1589398321",
            "testuser@testdom.com",
            None,
            "00000000000000000000000000000000",
            "8ADB9B997580D69E69CAA2BBB68F4697",
            CredentialCache("testuser@testdom.com"),
        ),
    ],
)
def test_get_credential_from_file(line, username, domain, lm_hash, nt_hash, input, tmpdir, monkeypatch):
    tmp_creds = os.path.join(to_text(tmpdir), "pÿspᴞӛgӫ TÈ$" ".creds")
    monkeypatch.setenv("NTLM_USER_FILE", tmp_creds)
    with open(tmp_creds, mode="wb") as fd:
        fd.write(to_bytes(line))

    actual = ntlm._NTLMCredential(input)
    assert actual.username == username
    assert actual.domain == domain
    assert actual.lm_hash == base64.b16decode(lm_hash)
    assert actual.nt_hash == base64.b16decode(nt_hash)


def test_get_credential_from_file_no_matches(tmpdir, monkeypatch):
    tmp_creds = os.path.join(to_text(tmpdir), "pÿspᴞӛgӫ TÈ$" ".creds")
    monkeypatch.setenv("NTLM_USER_FILE", tmp_creds)
    with open(tmp_creds, mode="wb") as fd:
        fd.write(b"domain:username:password")

    with pytest.raises(
        SpnegoError, match="Failed to find any matching credential in NTLM_USER_FILE " "credential store."
    ):
        ntlm._NTLMCredential(CredentialCache("fake\\username"))


@pytest.mark.parametrize("level", [-1, 6])
def test_invalid_lm_compat_level(level, monkeypatch):
    monkeypatch.setenv("LM_COMPAT_LEVEL", str(level))

    expected = "Invalid LM_COMPAT_LEVEL %s, must be between 0 and 5" % level
    with pytest.raises(SpnegoError, match=re.escape(expected)):
        ntlm.NTLMProxy("user", "pass")


def test_context_no_store_initiate():
    with pytest.raises(
        OperationNotAvailableError,
        match="No username or password was specified and the credential cache did not exist or contained no credentials",
    ):
        ntlm.NTLMProxy(CredentialCache(), usage="initiate")


def test_context_no_store_accept():
    with pytest.raises(
        OperationNotAvailableError,
        match="NTLM acceptor requires NTLM credential cache to be provided through the env var NTLM_USER_FILE set to a filepath",
    ):
        ntlm.NTLMProxy(CredentialCache(), usage="accept")


def test_iov_available():
    assert ntlm.NTLMProxy.iov_available() is True


def test_ntlm_invalid_usage():
    with pytest.raises(ValueError, match="Invalid usage 'test', must be initiate or accept"):
        ntlm.NTLMProxy("user", "pass", usage="test")


def test_ntlm_invalid_protocol():
    with pytest.raises(ValueError, match="Invalid protocol 'fake', must be ntlm, kerberos, negotiate, or credssp"):
        ntlm.NTLMProxy("user", "pass", protocol="fake")


def test_ntlm_query_message_sizes_fail():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(NoContextError, match="Cannot get message sizes until context has been established"):
        n.query_message_sizes()


def test_ntlm_wrap_qop_invalid():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(UnsupportedQop, match="Unsupported QoP value 1 specified for NTLM"):
        n.wrap(b"data", qop=1)


def test_ntlm_wrap_no_sign_or_seal():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(OperationNotAvailableError, match="NTLM wrap without integrity or confidentiality"):
        n.wrap(b"data")


def test_ntlm_wrap_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    with pytest.raises(NoContextError, match="Cannot wrap until context has been established"):
        n.wrap(b"data")


def test_ntlm_wrap_winrm_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    with pytest.raises(NoContextError, match="Cannot wrap until context has been established"):
        n.wrap_winrm(b"data")


def test_ntlm_unwrap_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(NoContextError, match="Cannot unwrap until context has been established"):
        n.unwrap(b"data")


def test_ntlm_unwrap_winrm_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(NoContextError, match="Cannot unwrap until context has been established"):
        n.unwrap_winrm(b"header", b"data")


def test_ntlm_wrap_iov_with_qop_fail():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(UnsupportedQop, match="Unsupported QoP value 1 specified for NTLM"):
        n.wrap_iov([], qop=1)


def test_ntlm_wrap_iov_no_sign_or_seal():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(OperationNotAvailableError, match="NTLM wrap without integrity or confidentiality"):
        n.wrap_iov([])


def test_ntlm_wrap_iov_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    with pytest.raises(NoContextError, match="Cannot wrap until context has been established"):
        n.wrap_iov([])


def test_ntlm_wrap_iov_no_header():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_out = n._handle_out = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="wrap_iov no IOV header buffer present"):
        n.wrap_iov([b""])


def test_ntlm_wrap_iov_no_data():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_out = n._handle_out = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="wrap_iov no IOV data buffer present"):
        n.wrap_iov([spnego.iov.BufferType.header])


def test_ntlm_wrap_iov_multiple_header():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_out = n._handle_out = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="wrap_iov must only be used with 1 header IOV buffer"):
        n.wrap_iov([spnego.iov.BufferType.header, b"", spnego.iov.BufferType.header])


def test_ntlm_wrap_iov_multiple_data():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_out = n._handle_out = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="wrap_iov must only be used with 1 data IOV buffer"):
        n.wrap_iov([spnego.iov.BufferType.header, b"", b""])


def test_ntlm_wrap_iov_invalid_type():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_out = n._handle_out = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="wrap_iov unsupported IOV buffer type padding"):
        n.wrap_iov([spnego.iov.BufferType.header, b"", spnego.iov.BufferType.padding])


def test_ntlm_wrap_iov_data_not_bytes():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_out = n._handle_out = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="wrap_iov IOV data buffer at \\[1\\] must be bytes"):
        n.wrap_iov([spnego.iov.BufferType.header, (spnego.iov.BufferType.data, 1)])


def test_ntlm_wrap_iov_signonly_not_bytes():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_out = n._handle_out = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="wrap_iov IOV sign_only buffer at \\[1\\] must be bytes"):
        n.wrap_iov([spnego.iov.BufferType.header, (spnego.iov.BufferType.sign_only, 1)])


def test_ntlm_unwrap_iov_no_sign_or_seal():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(OperationNotAvailableError, match="NTLM unwrap without integrity or confidentiality"):
        n.unwrap_iov([])


def test_ntlm_unwrap_iov_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    with pytest.raises(NoContextError, match="Cannot unwrap until context has been established"):
        n.unwrap_iov([])


def test_ntlm_unwrap_iov_no_header():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_in = n._handle_in = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="unwrap_iov no IOV header buffer present"):
        n.unwrap_iov([b""])


def test_ntlm_unwrap_iov_no_data():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_in = n._handle_in = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="unwrap_iov no IOV data buffer present"):
        n.unwrap_iov([(spnego.iov.BufferType.header, b"")])


def test_ntlm_unwrap_iov_multiple_header():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_in = n._handle_in = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="unwrap_iov must only be used with 1 header IOV buffer"):
        n.unwrap_iov([(spnego.iov.BufferType.header, b""), b"", (spnego.iov.BufferType.header, b"")])


def test_ntlm_unwrap_iov_multiple_data():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_in = n._handle_in = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="unwrap_iov must only be used with 1 data IOV buffer"):
        n.unwrap_iov([(spnego.iov.BufferType.header, b""), b"", b""])


def test_ntlm_unwrap_iov_invalid_type():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_in = n._handle_in = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="unwrap_iov unsupported IOV buffer type padding"):
        n.unwrap_iov([(spnego.iov.BufferType.header, b""), b"", (spnego.iov.BufferType.padding, b"")])


def test_ntlm_unwrap_iov_data_not_bytes():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_in = n._handle_in = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="unwrap_iov IOV data buffer at \\[1\\] must be bytes"):
        n.unwrap_iov([(spnego.iov.BufferType.header, b""), (spnego.iov.BufferType.data, 1)])


def test_ntlm_unwrap_iov_data_readonly_not_bytes():
    n = ntlm.NTLMProxy("user", "pass")
    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
    n._sign_key_in = n._handle_in = 1  # type: ignore

    with pytest.raises(InvalidTokenError, match="unwrap_iov IOV data_readonly buffer at \\[1\\] must be bytes"):
        n.unwrap_iov([(spnego.iov.BufferType.header, b""), (spnego.iov.BufferType.data_readonly, 1)])


def test_ntlm_sign_qop_invalid():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(UnsupportedQop, match="Unsupported QoP value 1 specified for NTLM"):
        n.sign(b"data", qop=1)


def test_ntlm_sign_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(NoContextError, match="Cannot sign until context has been established"):
        n.sign(b"data")


def test_ntlm_verify_no_context():
    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(NoContextError, match="Cannot verify until context has been established"):
        n.verify(b"data", b"mic")


def test_ntlm_no_encoding_flags():
    negotiate = memoryview(bytearray(get_data("ntlm_negotiate")))
    negotiate[12:16] = b"\x00\x00\x00\x00"

    n = ntlm.NTLMProxy("user", "pass")
    with pytest.raises(
        SpnegoError,
        match="Neither NEGOTIATE_OEM or NEGOTIATE_UNICODE flags were set, cannot derive " "encoding for text fields",
    ):
        n._step_accept_negotiate(negotiate.tobytes())


@pytest.mark.parametrize(
    "client_opt, present",
    [
        (spnego.NegotiateOptions.use_ntlm, False),
        (spnego.NegotiateOptions.use_ntlm, True),
        (spnego.NegotiateOptions.use_gssapi, True),
        (spnego.NegotiateOptions.use_sspi, True),
    ],
)
def test_ntlm_bad_bindings(client_opt, present, ntlm_cred):
    if client_opt & spnego.NegotiateOptions.use_gssapi:
        if "ntlm" not in spnego._gss.GSSAPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through GSSAPI")

    elif client_opt & spnego.NegotiateOptions.use_sspi:
        if "ntlm" not in spnego._sspi.SSPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through SSPI")

    initiator_cbt = None
    if present:
        initiator_cbt = spnego.channel_bindings.GssChannelBindings(application_data=b"tls-host-data:bad")

    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=client_opt,
        protocol="ntlm",
        channel_bindings=initiator_cbt,
    )

    acceptor_cbt = spnego.channel_bindings.GssChannelBindings(application_data=b"tls-host-data:test")
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm", channel_bindings=acceptor_cbt)

    auth = c.step(s.step(c.step()))

    if present:
        expected = "Acceptor bindings do not match initiator bindings"

    else:
        expected = "Acceptor bindings specified but not present in initiator response"

    with pytest.raises(BadBindingsError, match=expected):
        s.step(auth)


def test_ntlm_bad_mic(ntlm_cred):
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    auth = memoryview(bytearray(c.step(s.step(c.step())) or b""))
    auth[64:80] = b"\x01" * 16

    with pytest.raises(InvalidTokenError, match="Invalid MIC in NTLM authentication message"):
        s.step(auth.tobytes())


@pytest.mark.parametrize(
    "env_var, expected",
    [
        (None, to_text(socket.gethostname()).upper()),
        ("", None),
        ("custom", "custom"),
    ],
)
def test_ntlm_workstation_override(env_var, expected, ntlm_cred, monkeypatch):
    if env_var is not None:
        monkeypatch.setenv("NETBIOS_COMPUTER_NAME", env_var)

    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )

    b_negotiate = c.step()
    assert b_negotiate is not None
    negotiate = Negotiate.unpack(b_negotiate)

    flags = (
        negotiate.flags
        | NegotiateFlags.request_target
        | NegotiateFlags.ntlm
        | NegotiateFlags.always_sign
        | NegotiateFlags.target_info
        | NegotiateFlags.target_type_server
    )

    server_challenge = os.urandom(8)
    target_name = to_text(socket.gethostname()).upper()

    target_info = TargetInfo()
    target_info[AvId.nb_computer_name] = target_name
    target_info[AvId.nb_domain_name] = "WORKSTATION"
    target_info[AvId.dns_computer_name] = to_text(socket.getfqdn())
    target_info[AvId.timestamp] = FileTime.now()

    version = Version(10, 0, 0, 1)
    challenge = Challenge(flags, server_challenge, target_name=target_name, target_info=target_info, version=version)

    b_auth = c.step(challenge.pack())
    assert b_auth is not None
    auth = Authenticate.unpack(b_auth)

    assert auth.workstation == expected


@pytest.mark.parametrize(
    "include_time, expected",
    [
        # If the challenge didn't contain the time then the client should generate it's own otherwise it uses the challenge
        # time.
        (True, 0),
        (False, 1),
    ],
)
def test_ntlm_custom_time(include_time, expected, ntlm_cred, mocker, monkeypatch):
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )

    b_negotiate = c.step()
    assert b_negotiate is not None
    negotiate = Negotiate.unpack(b_negotiate)

    flags = (
        negotiate.flags
        | NegotiateFlags.request_target
        | NegotiateFlags.ntlm
        | NegotiateFlags.always_sign
        | NegotiateFlags.target_info
        | NegotiateFlags.target_type_server
    )

    server_challenge = os.urandom(8)
    target_name = to_text(socket.gethostname()).upper()

    target_info = TargetInfo()
    target_info[AvId.nb_computer_name] = target_name
    target_info[AvId.nb_domain_name] = "WORKSTATION"
    target_info[AvId.dns_computer_name] = to_text(socket.getfqdn())

    if include_time:
        target_info[AvId.timestamp] = FileTime.now()

    challenge = Challenge(flags, server_challenge, target_name=target_name, target_info=target_info)

    mock_now = mocker.MagicMock()
    mock_now.side_effect = FileTime.now
    monkeypatch.setattr(FileTime, "now", mock_now)

    c.step(challenge.pack())
    assert c.complete
    assert mock_now.call_count == expected


def test_ntlm_no_key_exch(ntlm_cred):
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    c._context_req &= ~0x40000000  # NTLMSSP_NEGOTIATE_KEY_EXCH

    auth = c.step(s.step(c.step()))
    s.step(auth)

    # Make sure EncryptedRandomSessionKeyFields was set to 0 (no KEY_EXCH).
    assert auth is not None
    assert auth[52:54] == b"\x00\x00"

    plaintext = os.urandom(32)

    c_wrap_result = c.wrap(plaintext)
    assert c_wrap_result.encrypted
    assert c_wrap_result.data != plaintext

    s_unwrap_result = s.unwrap(c_wrap_result.data)
    assert s_unwrap_result.data == plaintext
    assert s_unwrap_result.encrypted

    plaintext = os.urandom(17)

    s_wrap_result = s.wrap(plaintext)
    assert s_wrap_result.encrypted
    assert s_wrap_result.data != plaintext

    c_unwrap_result = c.unwrap(s_wrap_result.data)
    assert c_unwrap_result.data == plaintext
    assert c_unwrap_result.encrypted

    plaintext = os.urandom(3)
    c_sig = c.sign(plaintext)
    s.verify(plaintext, c_sig)

    plaintext = os.urandom(9)
    s_sig = s.sign(plaintext)
    c.verify(plaintext, s_sig)


def test_ntlm_lm_request(ntlm_cred, monkeypatch):
    monkeypatch.setenv("LM_COMPAT_LEVEL", "0")
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    auth = memoryview(bytearray(c.step(s.step(c.step())) or b""))
    auth[20:28] = b"\x00" * 8

    s.step(auth.tobytes())

    assert c.complete
    assert s.complete


def test_ntlm_no_lm_allowed(ntlm_cred, monkeypatch):
    monkeypatch.setenv("LM_COMPAT_LEVEL", "0")
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )

    monkeypatch.setenv("LM_COMPAT_LEVEL", "4")
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    auth = memoryview(bytearray(c.step(s.step(c.step())) or b""))
    auth[20:28] = b"\x00" * 8

    with pytest.raises(InvalidTokenError, match="Acceptor settings are set to reject LM responses"):
        s.step(auth)


def test_ntlm_nt_v1_request(ntlm_cred, monkeypatch):
    monkeypatch.setenv("LM_COMPAT_LEVEL", "0")
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )

    monkeypatch.setenv("LM_COMPAT_LEVEL", "4")
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    auth = c.step(s.step(c.step()))

    s.step(auth)

    assert c.complete
    assert s.complete


def test_ntlm_no_nt_v1_allowed(ntlm_cred, monkeypatch):
    monkeypatch.setenv("LM_COMPAT_LEVEL", "0")
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )

    monkeypatch.setenv("LM_COMPAT_LEVEL", "5")
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    auth = c.step(s.step(c.step()))

    with pytest.raises(InvalidTokenError, match="Acceptor settings are set to reject NTv1 responses"):
        s.step(auth)


def test_ntlm_with_invalid_surrogate_pair_pass(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: pathlib.Path,
) -> None:
    username = "user"

    # This is a password for a gMSA account used as a test. It is technically
    # a UTF-16-LE string value but contains invalid surrogate pairs which is
    # important for testing.
    b_gmsa_password = (
        b"\x91\x45\xC9\xD1\x1D\x74\xA9\xE2"
        b"\x27\x5A\x4C\xBE\x13\xC1\xE2\xF4"
        b"\x89\x94\x49\x5E\x01\x60\xDD\xBA"
        b"\xE2\xD9\x12\x53\xF0\xEB\x96\x38"
        b"\xB3\x8B\xD2\x17\xC9\xCC\x9B\xB6"
        b"\xC7\xF0\xCC\x8F\xEB\x75\x03\x77"
        b"\x30\xD3\xE2\x6C\xE6\x00\x04\x39"
        b"\xF5\x5F\xD9\xA5\xD8\xEF\xB3\x9F"
        b"\xDE\x4A\xB5\xC1\x51\xC0\x44\x3B"
        b"\x66\xC6\xF4\x68\x8D\xE1\x78\xBE"
        b"\x3D\x35\x34\xC7\x4A\x91\x6D\x7B"
        b"\x3C\xE0\x6E\x1C\xE9\xA9\x96\x6E"
        b"\xDA\x09\x6A\x39\x1A\x2E\x5F\xD2"
        b"\x92\x86\x46\x3B\x8B\x9A\xD3\xCE"
        b"\xED\x83\x03\x2A\x33\x0D\xBC\x06"
        b"\x91\xC6\x0C\xB1\x69\x5D\x2D\x59"
        b"\xE0\x66\x18\x99\x00\xD1\x5A\x55"
        b"\x85\xA3\xA8\x23\x0E\xCC\x16\x08"
        b"\xF9\xE1\x9D\xF7\x09\x24\x66\xB3"
        b"\x56\x6D\xC8\x2B\x4D\x33\x7F\x1A"
        b"\xED\x69\x24\x09\xB1\x0C\xD6\x51"
        b"\xBB\x62\xD9\x82\xD4\xA6\x1D\x91"
        b"\x6F\xC4\xB2\xB0\x45\x9A\x40\x5A"
        b"\xEC\x81\x71\xA1\x48\xB3\x52\x37"
        b"\x26\x72\x98\x01\x22\x31\xF2\xD8"
        b"\xD4\x83\x7B\xF3\xCA\xD5\x81\x24"
        b"\xDC\xA9\xC2\xBF\x6D\x8E\x87\x7D"
        b"\x24\x87\x49\x6C\x46\xE6\x67\x8B"
        b"\x10\x69\x00\x04\xCA\x17\x4B\xC8"
        b"\x04\x33\x69\x06\x61\x57\xB9\xC7"
        b"\x3B\xFC\x0A\xCD\x35\xCE\x61\xB9"
        b"\x87\x3B\xFF\x3A\x2D\x55\x67\xF6"
    )

    tmp_creds = tmp_path / "ntlm.cred"
    with open(tmp_creds, mode="w") as fd:
        nt_hash = base64.b16encode(md4(b_gmsa_password)).decode()
        fd.write(f"{username}:1:00000000000000000000000000000000:{nt_hash}:[U]:LCT-1589398321")

    monkeypatch.setenv("NTLM_USER_FILE", str(tmp_creds.absolute()))

    c = spnego.client(
        username,
        b_gmsa_password.decode("utf-16-le", errors="surrogatepass"),
        hostname=socket.gethostname(),
        options=spnego.NegotiateOptions.use_ntlm,
        protocol="ntlm",
    )
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    s.step(c.step(s.step(c.step())))

    assert c.complete
    assert s.complete


@pytest.mark.parametrize(
    "client_opt",
    [
        spnego.NegotiateOptions.use_ntlm,
        spnego.NegotiateOptions.use_gssapi,
        spnego.NegotiateOptions.use_sspi,
    ],
)
def test_ntlm_invalid_password(client_opt, ntlm_cred):
    if client_opt & spnego.NegotiateOptions.use_gssapi:
        if "ntlm" not in spnego._gss.GSSAPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through GSSAPI")

    elif client_opt & spnego.NegotiateOptions.use_sspi:
        if "ntlm" not in spnego._sspi.SSPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through SSPI")

    c = spnego.client(ntlm_cred[0], "Invalid", hostname=socket.gethostname(), options=client_opt, protocol="ntlm")
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    auth = c.step(s.step(c.step()))

    with pytest.raises(InvalidTokenError, match="Invalid NTLM response from initiator"):
        s.step(auth)


@pytest.mark.parametrize(
    "client_opt",
    [
        spnego.NegotiateOptions.use_ntlm,
        spnego.NegotiateOptions.use_gssapi,
        spnego.NegotiateOptions.use_sspi,
    ],
)
def test_ntlm_verify_fail(client_opt, ntlm_cred):
    if client_opt & spnego.NegotiateOptions.use_gssapi:
        if "ntlm" not in spnego._gss.GSSAPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through GSSAPI")

    elif client_opt & spnego.NegotiateOptions.use_sspi:
        if "ntlm" not in spnego._sspi.SSPIProxy.available_protocols():
            pytest.skip("Test requires NTLM to be available through SSPI")

    c = spnego.client(ntlm_cred[0], ntlm_cred[1], hostname=socket.gethostname(), options=client_opt, protocol="ntlm")
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    s.step(c.step(s.step(c.step())))

    c.sign(b"data")
    sig = c.sign(b"data 2")

    with pytest.raises(BadMICError, match="Invalid Message integrity Check"):
        s.verify(b"data", sig)


def test_ntlm_anon_response(ntlm_cred):
    c = spnego.client(ntlm_cred[0], ntlm_cred[1], options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")
    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")

    auth = Authenticate.unpack(c.step(s.step(c.step())) or b"")
    anon_auth = Authenticate(flags=auth.flags, lm_challenge_response=b"\x00", nt_challenge_response=b"").pack()

    with pytest.raises(OperationNotAvailableError, match="Anonymous user authentication not implemented"):
        s.step(anon_auth)


def test_ntlm_iov_wrapping(ntlm_cred):
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        protocol="ntlm",
        options=spnego.NegotiateOptions.use_ntlm,
    )
    s = spnego.server(protocol="ntlm")

    s.step(c.step(s.step(c.step())))

    res1 = c.wrap_iov(
        [
            (spnego.iov.BufferType.sign_only, b"sign 1"),
            b"data",
            (spnego.iov.BufferType.data_readonly, b"sign 2"),
            spnego.iov.BufferType.header,
        ]
    )
    assert isinstance(res1, spnego.IOVWrapResult)
    assert res1.encrypted
    assert len(res1.buffers) == 4
    assert res1.buffers[0].data == b"sign 1"
    assert res1.buffers[1].data != b"data"
    assert res1.buffers[2].data == b"sign 2"
    assert len(res1.buffers[3].data or b"") == 16

    res2 = s.unwrap_iov(res1.buffers)
    assert isinstance(res2, spnego.IOVUnwrapResult)
    assert res2.encrypted
    assert res2.qop == 0
    assert len(res2.buffers) == 4
    assert res2.buffers[0].data == b"sign 1"
    assert res2.buffers[1].data == b"data"
    assert res2.buffers[2].data == b"sign 2"
    assert res2.buffers[3].data == res1.buffers[3].data

    res3 = s.wrap_iov(
        [
            (spnego.iov.BufferType.sign_only, b"sign 1"),
            b"data",
            (spnego.iov.BufferType.data_readonly, b"sign 2"),
            spnego.iov.BufferType.header,
        ]
    )
    assert isinstance(res3, spnego.IOVWrapResult)
    assert res3.encrypted
    assert len(res3.buffers) == 4
    assert res3.buffers[0].data == b"sign 1"
    assert res3.buffers[1].data != b"data"
    assert res3.buffers[2].data == b"sign 2"
    assert len(res3.buffers[3].data or b"") == 16

    res4 = c.unwrap_iov(res3.buffers)
    assert isinstance(res4, spnego.IOVUnwrapResult)
    assert res4.encrypted
    assert res4.qop == 0
    assert len(res4.buffers) == 4
    assert res4.buffers[0].data == b"sign 1"
    assert res4.buffers[1].data == b"data"
    assert res4.buffers[2].data == b"sign 2"
    assert res4.buffers[3].data == res3.buffers[3].data


def test_ntlm_iov_unwrapping_as_stream(ntlm_cred):
    c = spnego.client(
        ntlm_cred[0],
        ntlm_cred[1],
        hostname=socket.gethostname(),
        protocol="ntlm",
        options=spnego.NegotiateOptions.use_ntlm,
    )
    s = spnego.server(protocol="ntlm")

    s.step(c.step(s.step(c.step())))

    res1 = c.wrap_iov(
        [
            b"data",
            spnego.iov.BufferType.header,
        ]
    )
    assert isinstance(res1, spnego.IOVWrapResult)
    assert res1.encrypted
    assert len(res1.buffers) == 2
    assert res1.buffers[0].data != b"data"
    assert len(res1.buffers[1].data or b"") == 16

    msg = (res1.buffers[1].data or b"") + (res1.buffers[0].data or b"")
    res2 = s.unwrap_iov(
        [
            (spnego.iov.BufferType.stream, msg),
            spnego.iov.BufferType.data,
        ]
    )
    assert isinstance(res2, spnego.IOVUnwrapResult)
    assert res2.encrypted
    assert res2.qop == 0
    assert len(res2.buffers) == 2
    assert res2.buffers[1].data == b"data"

    res3 = s.wrap_iov(
        [
            spnego.iov.BufferType.header,
            b"data",
        ]
    )
    assert isinstance(res3, spnego.IOVWrapResult)
    assert res3.encrypted
    assert len(res3.buffers) == 2
    assert len(res3.buffers[0].data or b"") == 16
    assert res3.buffers[1].data != b"data"

    msg = (res3.buffers[0].data or b"") + (res3.buffers[1].data or b"")
    res4 = c.unwrap_iov(
        [
            (spnego.iov.BufferType.stream, msg),
            spnego.iov.BufferType.data,
        ]
    )
    assert isinstance(res4, spnego.IOVUnwrapResult)
    assert res4.encrypted
    assert res4.qop == 0
    assert len(res4.buffers) == 2
    assert res4.buffers[1].data == b"data"



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import re
import socket

import pytest

import spnego
import spnego._credssp as credssp
from spnego._credssp_structures import NegoData, TSRequest
from spnego.exceptions import (
    BadBindingsError,
    FeatureMissingError,
    InvalidTokenError,
    NegotiateOptions,
    NoContextError,
    OperationNotAvailableError,
    SpnegoError,
)
from spnego.tls import default_tls_context


@pytest.mark.parametrize(
    "expected, usage, nonce",
    [
        (b"\x00\x01", "initiate", None),
        (
            b"\x45\x90\xC2\x6C\x83\x7D\xAC\xDF\x3D\xA8\xFA\xE5\x78\x26\x4D\x9E"
            b"\x45\x48\x6A\xFC\x39\x51\xF2\x1E\xF3\x57\x06\x4C\xD5\x5F\xBB\xB5",
            "initiate",
            b"\x00\x00",
        ),
        (b"\x01\x01", "accept", None),
        (
            b"\xCD\xF4\xFF\xEE\x6C\xB6\x39\x5E\x5B\x69\x2C\x2D\xA3\x97\x38\x6F"
            b"\x6A\x1E\xAB\x2D\x3A\x62\x71\x81\x6A\x04\xC7\xA2\x91\x4D\x33\x3F",
            "accept",
            b"\x00\x00",
        ),
    ],
)
def test_pub_key_auth(expected, usage, nonce):
    actual = credssp._get_pub_key_auth(b"\x00\x01", usage, nonce=nonce)

    assert actual == expected


@pytest.mark.parametrize(
    "protocol, cipher, expected",
    [
        ["TLSv1.3", "TLS_AES_128_GCM_SHA256", 17],
        ["TLSv1.3", "TLS_AES_256_GCM_SHA384", 17],
        ["TLSv1.2", "ECDHE-RSA-AES128-GCM-SHA256", 16],
        ["TLSv1.2", "RC4-MD5", 16],
        ["TLSv1.2", "ECDH-ECDSA-3DES-SHA256", 34],
        ["TLSv1.2", "ECDH-RSA-AES-SHA384", 50],
        ["TLSv1.2", "ECDH-RSA-AES", 2],
    ],
)
def test_get_credssp_trailer_length(protocol, cipher, expected):
    actual = credssp._tls_trailer_length(30, protocol, cipher)
    assert actual == expected


def test_credssp_no_iov():
    assert not credssp.CredSSPProxy("username", "password").iov_available()


def test_credssp_iov_not_available():
    expected = (
        "The system is missing the GSSAPI IOV extension headers or CredSSP is being requested, "
        "cannot utilize wrap_iov and unwrap_iov"
    )
    with pytest.raises(FeatureMissingError, match=re.escape(expected)):
        credssp.CredSSPProxy("username", "password", options=spnego.NegotiateOptions.wrapping_iov)


def test_credssp_no_session_key():
    with pytest.raises(FeatureMissingError, match="The protocol selected does not support getting the session key"):
        credssp.CredSSPProxy("user", "password", options=NegotiateOptions.session_key)


def test_credssp_fail_getting_session_key():
    context = credssp.CredSSPProxy("username", "password")

    with pytest.raises(OperationNotAvailableError, match="CredSSP does not have a session key to share"):
        _ = context.session_key


def test_credssp_wrap_iov_fail():
    with pytest.raises(OperationNotAvailableError, match="CredSSP does not offer IOV wrapping"):
        credssp.CredSSPProxy("username", "password").wrap_iov([])


def test_credssp_unwrap_iov_fail():
    with pytest.raises(OperationNotAvailableError, match="CredSSP does not offer IOV wrapping"):
        credssp.CredSSPProxy("username", "password").unwrap_iov([])


def test_credssp_sign_fail():
    with pytest.raises(OperationNotAvailableError, match="CredSSP does not offer signing"):
        credssp.CredSSPProxy("username", "password").sign(b"")


def test_credssp_verify_fail():
    with pytest.raises(OperationNotAvailableError, match="CredSSP does not offer verification"):
        credssp.CredSSPProxy("username", "password").verify(b"", b"")


def test_credssp_wrap_no_context():
    with pytest.raises(SpnegoError, match="Invalid TLS state when wrapping data"):
        credssp.CredSSPProxy("username", "password").wrap(b"data")


def test_credssp_query_message_sizes_fail():
    with pytest.raises(NoContextError, match="Cannot get message sizes until context has been established"):
        credssp.CredSSPProxy("username", "password").query_message_sizes()


def test_credssp_invalid_handshake(ntlm_cred):
    c = credssp.CredSSPProxy(ntlm_cred[0], ntlm_cred[1], protocol="credssp")
    s = credssp.CredSSPProxy(None, None, protocol="credssp", usage="accept")

    server_hello = s.step(c.step())
    assert server_hello is not None

    with pytest.raises(InvalidTokenError, match="TLS handshake for CredSSP"):
        c.step(b"\x00" + server_hello)


def test_credssp_server_without_pub_key():
    context = default_tls_context(usage="accept")
    with pytest.raises(OperationNotAvailableError, match="Provided tls context does not have a public key set"):
        credssp.CredSSPProxy("username", "password", usage="accept", credssp_tls_context=context)


@pytest.mark.parametrize("version", [2, 5])
def test_credssp_invalid_client_authentication_v2(version, ntlm_cred, monkeypatch):
    monkeypatch.setattr(credssp, "_CREDSSP_VERSION", version)

    c = credssp.CredSSPProxy(ntlm_cred[0], ntlm_cred[1], hostname=socket.getfqdn(), protocol="credssp")
    s = credssp.CredSSPProxy(None, None, hostname=socket.getfqdn(), protocol="credssp", usage="accept")

    # Set up the state so the server can send the error code
    server_tls_token = None
    while c._auth_context is None:
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    bad_request = TSRequest(credssp._CREDSSP_VERSION, nego_tokens=NegoData(b"\x00"))
    actual = s.step(c.wrap(bad_request.pack()).data)

    # On CredSSP v2 or 5 there is no error message from the server, just no token
    assert actual is None


@pytest.mark.parametrize("version", [3, 4, 6])
def test_credssp_invalid_client_authentication(version, ntlm_cred, monkeypatch):
    monkeypatch.setattr(credssp, "_CREDSSP_VERSION", version)

    c = credssp.CredSSPProxy(ntlm_cred[0], ntlm_cred[1], hostname=socket.getfqdn(), protocol="credssp")
    s = credssp.CredSSPProxy(None, None, hostname=socket.getfqdn(), protocol="credssp", usage="accept")

    # Set up the state so the server can send the error code
    server_tls_token = None
    while c._auth_context is None:
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    assert server_tls_token is not None
    c.unwrap(server_tls_token)
    bad_request = TSRequest(credssp._CREDSSP_VERSION, nego_tokens=NegoData(b"\x00"))
    error_msg = s.step(c.wrap(bad_request.pack()).data)

    expected = re.escape("Received NTStatus in TSRequest from acceptor, Context: Authentication")
    with pytest.raises(InvalidTokenError, match=expected):
        c.step(error_msg)


def test_credssp_no_pub_key_after_auth(ntlm_cred):
    c = credssp.CredSSPProxy(ntlm_cred[0], ntlm_cred[1], hostname=socket.getfqdn(), protocol="credssp")
    s = credssp.CredSSPProxy(None, None, hostname=socket.getfqdn(), protocol="credssp", usage="accept")

    server_tls_token = None
    while c._auth_context is None:
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    ntlm3_pub_key = c.step(server_tls_token)

    # Send back a TSRequest without the public key.
    assert ntlm3_pub_key is not None
    s.unwrap(ntlm3_pub_key)
    ts_request = TSRequest(credssp._CREDSSP_VERSION)

    with pytest.raises(InvalidTokenError, match="Acceptor did not response with pubKeyAuth info"):
        c.step(s.wrap(ts_request.pack()).data)


def test_credssp_pub_key_mismatch_initiator(ntlm_cred):
    options = NegotiateOptions.use_ntlm
    c = credssp.CredSSPProxy(ntlm_cred[0], ntlm_cred[1], hostname=socket.getfqdn(), protocol="credssp", options=options)
    s = credssp.CredSSPProxy(None, None, hostname=socket.getfqdn(), protocol="credssp", usage="accept", options=options)

    server_tls_token = None
    while c._auth_context is None:
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    ntlm3_pub_key = c.step(server_tls_token)

    s_auth_context = s._auth_context
    assert s_auth_context is not None

    # Send back a TSRequest with a bad pub_key_auth value.
    assert ntlm3_pub_key is not None
    request = TSRequest.unpack(s.unwrap(ntlm3_pub_key).data)
    assert request.nego_tokens is not None
    s_auth_context.step(request.nego_tokens[0].nego_token)
    ts_request = TSRequest(credssp._CREDSSP_VERSION, pub_key_auth=s_auth_context.wrap(b"bad").data)

    with pytest.raises(BadBindingsError, match="Public key verification failed, potential man in the middle attack"):
        c.step(s.wrap(ts_request.pack()).data)


def test_credssp_pub_key_mismatch_acceptor(ntlm_cred):
    options = NegotiateOptions.use_ntlm
    c = credssp.CredSSPProxy(ntlm_cred[0], ntlm_cred[1], hostname=socket.getfqdn(), protocol="credssp", options=options)
    s = credssp.CredSSPProxy(None, None, hostname=socket.getfqdn(), protocol="credssp", usage="accept", options=options)

    server_tls_token = None
    while c._auth_context is None:
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    # Craft the TSRequest with the NTLMv3 token and invalid pub_key_auth
    assert server_tls_token is not None
    ntlm2_request = TSRequest.unpack(c.unwrap(server_tls_token).data)
    assert ntlm2_request.nego_tokens is not None
    ntlm3 = c._auth_context.step(ntlm2_request.nego_tokens[0].nego_token)
    assert ntlm3 is not None

    request = TSRequest(
        credssp._CREDSSP_VERSION, nego_tokens=NegoData(ntlm3), pub_key_auth=c._auth_context.wrap(b"bad").data
    )

    with pytest.raises(BadBindingsError, match="Public key verification failed, potential man in the middle attack"):
        s.step(c.wrap(request.pack()).data)


def test_credssp_no_credential(ntlm_cred):
    options = NegotiateOptions.use_ntlm
    c = credssp.CredSSPProxy(ntlm_cred[0], ntlm_cred[1], hostname=socket.getfqdn(), protocol="credssp", options=options)
    s = credssp.CredSSPProxy(None, None, hostname=socket.getfqdn(), protocol="credssp", usage="accept", options=options)

    server_tls_token = None
    while c._auth_context is None:
        client_tls_token = c.step(server_tls_token)
        assert not c.complete
        assert not s.complete

        server_tls_token = s.step(client_tls_token)
        assert not c.complete
        assert not s.complete

    ntlm3_pub_key = c.step(server_tls_token)
    assert ntlm3_pub_key is not None

    pub_key_resp = s.step(ntlm3_pub_key)
    assert pub_key_resp is not None
    server_pub_key = TSRequest.unpack(c.unwrap(pub_key_resp).data)

    # Unpack the TSRequest properly and craft a TSRequest without a credential
    assert server_pub_key.pub_key_auth is not None
    c._auth_context.unwrap(server_pub_key.pub_key_auth)

    request = TSRequest(credssp._CREDSSP_VERSION)

    with pytest.raises(InvalidTokenError, match="No credential received on CredSSP TSRequest from initiator"):
        s.step(c.wrap(request.pack()).data)



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pytest

import spnego._credssp_structures as credssp

from .conftest import get_data


def test_unpack_missing_text_field():
    with pytest.raises(ValueError, match="Missing mandatory text field 'field' in 'structure'"):
        credssp.unpack_text_field({}, 0, "structure", "field")


def test_ts_request_pack():
    expected = get_data("credssp_ts_request")

    actual = credssp.TSRequest(6, credssp.NegoData(b"123"), b"auth info", b"pub key auth", 10, b"client nonce").pack()

    assert actual == expected


def test_ts_request_unpack():
    data = get_data("credssp_ts_request")

    actual = credssp.TSRequest.unpack(data)

    assert isinstance(actual, credssp.TSRequest)
    assert actual.version == 6
    assert isinstance(actual.nego_tokens, list)
    assert len(actual.nego_tokens) == 1
    assert isinstance(actual.nego_tokens[0], credssp.NegoData)
    assert actual.nego_tokens[0].nego_token == b"123"
    assert actual.auth_info == b"auth info"
    assert actual.pub_key_auth == b"pub key auth"
    assert actual.error_code == 10
    assert actual.client_nonce == b"client nonce"


def test_ts_credential_unknown_credential_type():
    credential = credssp.TSCredentials("failure")  # type: ignore[arg-type] # Testing this scenario

    with pytest.raises(ValueError, match="Invalid credential type set"):
        _ = credential.cred_type


def test_ts_credential_unknown_credential_type_unpack():
    data = get_data("credssp_ts_credential_password")
    # Manually change the credType to 0 in a known good structure.
    data = data[:6] + b"\x00" + data[7:]

    with pytest.raises(ValueError, match="Unknown credType 0 in TSCredentials, cannot unpack"):
        credssp.TSCredentials.unpack(data)


def test_ts_credential_password_pack():
    expected = get_data("credssp_ts_credential_password")
    actual = credssp.TSCredentials(
        credssp.TSPasswordCreds(
            "domain name",
            "username",
            "password",
        )
    ).pack()

    assert actual == expected


def test_ts_credential_password_unpack():
    data = get_data("credssp_ts_credential_password")
    actual = credssp.TSCredentials.unpack(data)

    assert isinstance(actual, credssp.TSCredentials)
    assert actual.cred_type == 1
    assert isinstance(actual.credentials, credssp.TSPasswordCreds)
    assert actual.credentials.domain_name == "domain name"
    assert actual.credentials.username == "username"
    assert actual.credentials.password == "password"


def test_ts_credential_smart_card_pack():
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/94846575-5a58-44de-b07b-48b90af328fb
    expected = get_data("credssp_ts_credential_smart_card")

    actual = credssp.TSCredentials(
        credssp.TSSmartCardCreds(
            "bbbbbbbbbbbb",
            credssp.TSCspDataDetail(
                1,
                reader_name="OMNIKEY CardMan 3x21 0",
                container_name="le-MSSmartcardUser-8bda019f-1266--53268",
                csp_name="Microsoft Base Smart Card Crypto Provider",
            ),
        )
    ).pack()

    assert actual == expected


def test_ts_credential_smart_card_unpack():
    # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/94846575-5a58-44de-b07b-48b90af328fb
    data = get_data("credssp_ts_credential_smart_card")
    actual = credssp.TSCredentials.unpack(data)

    assert isinstance(actual, credssp.TSCredentials)
    assert actual.cred_type == 2
    assert isinstance(actual.credentials, credssp.TSSmartCardCreds)
    assert actual.credentials.pin == "bbbbbbbbbbbb"
    assert isinstance(actual.credentials.csp_data, credssp.TSCspDataDetail)
    assert actual.credentials.csp_data.key_spec == 1
    assert actual.credentials.csp_data.card_name is None
    assert actual.credentials.csp_data.reader_name == "OMNIKEY CardMan 3x21 0"
    assert actual.credentials.csp_data.container_name == "le-MSSmartcardUser-8bda019f-1266--53268"
    assert actual.credentials.csp_data.csp_name == "Microsoft Base Smart Card Crypto Provider"
    assert actual.credentials.user_hint is None
    assert actual.credentials.domain_hint is None


def test_ts_credential_smart_card_full_pack():
    expected = get_data("credssp_ts_credential_smart_card_full")

    actual = credssp.TSCredentials(
        credssp.TSSmartCardCreds(
            "bbbbbbbbbbbb",
            credssp.TSCspDataDetail(
                1,
                card_name="Card Name",
                reader_name="OMNIKEY CardMan 3x21 0",
                container_name="le-MSSmartcardUser-8bda019f-1266--53268",
                csp_name="Microsoft Base Smart Card Crypto Provider",
            ),
            "user_hint",
            "domain_hint",
        )
    ).pack()

    assert actual == expected


def test_ts_credential_smart_card_full_unpack():
    data = get_data("credssp_ts_credential_smart_card_full")
    actual = credssp.TSCredentials.unpack(data)

    assert isinstance(actual, credssp.TSCredentials)
    assert actual.cred_type == 2
    assert isinstance(actual.credentials, credssp.TSSmartCardCreds)
    assert actual.credentials.pin == "bbbbbbbbbbbb"
    assert isinstance(actual.credentials.csp_data, credssp.TSCspDataDetail)
    assert actual.credentials.csp_data.key_spec == 1
    assert actual.credentials.csp_data.card_name == "Card Name"
    assert actual.credentials.csp_data.reader_name == "OMNIKEY CardMan 3x21 0"
    assert actual.credentials.csp_data.container_name == "le-MSSmartcardUser-8bda019f-1266--53268"
    assert actual.credentials.csp_data.csp_name == "Microsoft Base Smart Card Crypto Provider"
    assert actual.credentials.user_hint == "user_hint"
    assert actual.credentials.domain_hint == "domain_hint"


def test_ts_credential_remote_guard_multiple_pack():
    expected = get_data("credssp_ts_credential_remote_guard_multiple")

    actual = credssp.TSCredentials(
        credssp.TSRemoteGuardCreds(
            credssp.TSRemoteGuardPackageCred("Package 1", b"123"),
            [
                credssp.TSRemoteGuardPackageCred("Package 2", b"456"),
                credssp.TSRemoteGuardPackageCred("Package 3", b"789"),
            ],
        )
    ).pack()

    assert actual == expected


def test_ts_credential_remote_guard_multiple_unpack():
    data = get_data("credssp_ts_credential_remote_guard_multiple")
    actual = credssp.TSCredentials.unpack(data)

    assert isinstance(actual, credssp.TSCredentials)
    assert actual.cred_type == 6
    assert isinstance(actual.credentials, credssp.TSRemoteGuardCreds)

    assert isinstance(actual.credentials, credssp.TSRemoteGuardCreds)
    assert isinstance(actual.credentials.logon_cred, credssp.TSRemoteGuardPackageCred)
    assert actual.credentials.logon_cred.package_name == "Package 1"
    assert actual.credentials.logon_cred.cred_buffer == b"123"
    assert isinstance(actual.credentials.supplemental_creds, list)
    assert len(actual.credentials.supplemental_creds) == 2
    assert isinstance(actual.credentials.supplemental_creds[0], credssp.TSRemoteGuardPackageCred)
    assert actual.credentials.supplemental_creds[0].package_name == "Package 2"
    assert actual.credentials.supplemental_creds[0].cred_buffer == b"456"
    assert isinstance(actual.credentials.supplemental_creds[1], credssp.TSRemoteGuardPackageCred)
    assert actual.credentials.supplemental_creds[1].package_name == "Package 3"
    assert actual.credentials.supplemental_creds[1].cred_buffer == b"789"


def test_ts_credential_remote_guard_no_supplemental_pack():
    expected = get_data("credssp_ts_credential_remote_guard_no_supplemental")

    actual = credssp.TSCredentials(
        credssp.TSRemoteGuardCreds(
            credssp.TSRemoteGuardPackageCred("Package 1", b"123"),
        )
    ).pack()

    assert actual == expected


def test_ts_credential_remote_guard_no_supplemental_unpack():
    data = get_data("credssp_ts_credential_remote_guard_no_supplemental")
    actual = credssp.TSCredentials.unpack(data)

    assert isinstance(actual, credssp.TSCredentials)
    assert actual.cred_type == 6
    assert isinstance(actual.credentials, credssp.TSRemoteGuardCreds)

    assert isinstance(actual.credentials, credssp.TSRemoteGuardCreds)
    assert isinstance(actual.credentials.logon_cred, credssp.TSRemoteGuardPackageCred)
    assert actual.credentials.logon_cred.package_name == "Package 1"
    assert actual.credentials.logon_cred.cred_buffer == b"123"
    assert actual.credentials.supplemental_creds is None


def test_ts_credential_remote_guard_empty_supplemental_pack():
    expected = get_data("credssp_ts_credential_remote_guard_empty_supplemental")

    actual = credssp.TSCredentials(
        credssp.TSRemoteGuardCreds(credssp.TSRemoteGuardPackageCred("Package 1", b"123"), [])
    ).pack()

    assert actual == expected


def test_ts_credential_remote_guard_empty_supplemental_unpack():
    data = get_data("credssp_ts_credential_remote_guard_empty_supplemental")
    actual = credssp.TSCredentials.unpack(data)

    assert isinstance(actual, credssp.TSCredentials)
    assert actual.cred_type == 6
    assert isinstance(actual.credentials, credssp.TSRemoteGuardCreds)

    assert isinstance(actual.credentials, credssp.TSRemoteGuardCreds)
    assert isinstance(actual.credentials.logon_cred, credssp.TSRemoteGuardPackageCred)
    assert actual.credentials.logon_cred.package_name == "Package 1"
    assert actual.credentials.logon_cred.cred_buffer == b"123"
    assert actual.credentials.supplemental_creds == []


def test_ts_remote_guard_pack():
    # Based on https://interopevents.blob.core.windows.net/events/2018/rdp/day3/577741-Remote%20Credencial%20Guard.pdf
    expected = get_data("credssp_ts_remote_guard_ms_example")

    actual = credssp.TSRemoteGuardCreds(
        credssp.TSRemoteGuardPackageCred("Kerberos", b"\x11" * 2611),
        credssp.TSRemoteGuardPackageCred("NTLM", b"\x22" * 204),
    ).pack()

    assert actual == expected


def test_ts_remote_guard_unpack():
    # Based on https://interopevents.blob.core.windows.net/events/2018/rdp/day3/577741-Remote%20Credencial%20Guard.pdf
    data = get_data("credssp_ts_remote_guard_ms_example")
    actual = credssp.TSRemoteGuardCreds.unpack(data)

    assert isinstance(actual, credssp.TSRemoteGuardCreds)
    assert isinstance(actual.logon_cred, credssp.TSRemoteGuardPackageCred)
    assert actual.logon_cred.package_name == "Kerberos"
    assert actual.logon_cred.cred_buffer == b"\x11" * 2611
    assert isinstance(actual.supplemental_creds, list)
    assert len(actual.supplemental_creds) == 1
    assert isinstance(actual.supplemental_creds[0], credssp.TSRemoteGuardPackageCred)
    assert actual.supplemental_creds[0].package_name == "NTLM"
    assert actual.supplemental_creds[0].cred_buffer == b"\x22" * 204



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import datetime
import re
import typing

import pytest

import spnego._asn1 as asn1

ASN1_TAG_TESTS = [
    # Simple universal
    (asn1.TagClass.universal, False, asn1.TypeTagNumber.octet_string, b"\x00", b"\x04\x01\x00"),
    # Constructed value
    (asn1.TagClass.universal, True, asn1.TypeTagNumber.octet_string, b"\x00\x00", b"\x24\x02\x00\x00"),
    # Large tag number
    (asn1.TagClass.application, True, 1024, b"\x00\x00", b"\x7F\x88\x00\x02\x00\x00"),
    (asn1.TagClass.application, True, 1048576, b"\x00\x00", b"\x7F\xC0\x80\x00\x02\x00\x00"),
    # Long length
    (asn1.TagClass.universal, False, asn1.TypeTagNumber.octet_string, b"\x00" * 127, b"\x04\x7F" + (b"\x00" * 127)),
    (asn1.TagClass.universal, False, asn1.TypeTagNumber.octet_string, b"\x00" * 128, b"\x04\x81\x80" + (b"\x00" * 128)),
    (
        asn1.TagClass.universal,
        False,
        asn1.TypeTagNumber.octet_string,
        b"\x00" * 1024,
        b"\x04\x82\x04\x00" + (b"\x00" * 1024),
    ),
]

# openssl asn1parse -genstr 'INTEGER:<val>' -out test && hexdump -C test && rm test
INTEGER_TESTS = [  # INTEGER has weird rules that I don't fully understand, use a test of test cases.
    (-748591, b"\x02\x03\xF4\x93\xD1"),
    (-32769, b"\x02\x03\xFF\x7F\xFF"),
    (-32768, b"\x02\x02\x80\x00"),
    (-32767, b"\x02\x02\x80\x01"),
    (-257, b"\x02\x02\xFE\xFF"),
    (-256, b"\x02\x02\xFF\x00"),
    (-255, b"\x02\x02\xFF\x01"),
    (-129, b"\x02\x02\xFF\x7F"),
    (-128, b"\x02\x01\x80"),
    (-127, b"\x02\x01\x81"),
    (-17, b"\x02\x01\xEF"),
    (-16, b"\x02\x01\xF0"),
    (-10, b"\x02\x01\xF6"),
    (-1, b"\x02\x01\xFF"),
    (0, b"\x02\x01\x00"),
    (1, b"\x02\x01\x01"),
    (10, b"\x02\x01\x0A"),
    (16, b"\x02\x01\x10"),
    (17, b"\x02\x01\x11"),
    (127, b"\x02\x01\x7F"),
    (128, b"\x02\x02\x00\x80"),
    (129, b"\x02\x02\x00\x81"),
    (255, b"\x02\x02\x00\xFF"),
    (256, b"\x02\x02\x01\x00"),
    (257, b"\x02\x02\x01\x01"),
    (32767, b"\x02\x02\x7F\xFF"),
    (32768, b"\x02\x03\x00\x80\x00"),
    (32769, b"\x02\x03\x00\x80\x01"),
    (748591, b"\x02\x03\x0B\x6C\x2F"),
]

OID_TESTS = [
    ("1.2", b"\x06\x01\x2A"),
    ("1.2.3", b"\x06\x02\x2A\x03"),
    ("1.2.3.1024.2", b"\x06\x05\x2A\x03\x88\x00\x02"),
]


def test_tag_class_native_labels():
    actual = asn1.TagClass.native_labels()

    assert isinstance(actual, dict)
    assert actual[asn1.TagClass.universal] == "Universal"


def test_type_tag_number_native_labels():
    actual = asn1.TypeTagNumber.native_labels()

    assert isinstance(actual, dict)
    assert actual[asn1.TypeTagNumber.end_of_content] == "End-of-Content (EOC)"


@pytest.mark.parametrize(
    "value, tag_class, tag_number, expected",
    [
        (
            asn1.unpack_asn1(asn1.pack_asn1_octet_string(b"\x00\x01"))[0],
            asn1.TagClass.universal,
            asn1.TypeTagNumber.octet_string,
            b"\x00\x01",
        ),
        (b"\x00\x01", asn1.TagClass.universal, asn1.TypeTagNumber.octet_string, b"\x00\x01"),
        (asn1.ASN1Value(asn1.TagClass.application, True, 1023, b"\x00"), asn1.TagClass.application, 1023, b"\x00"),
    ],
)
def test_extract_asn1_tlv(value, tag_class, tag_number, expected):
    actual = asn1.extract_asn1_tlv(value, tag_class, tag_number)
    assert actual == expected


def test_extract_asn1_tlv_invalid_universal_class():
    expected = (
        "Invalid ASN.1 OCTET STRING tags, actual tag class TagClass.universal and tag number " "TypeTagNumber.integer"
    )
    with pytest.raises(ValueError, match=re.escape(expected)):
        asn1.extract_asn1_tlv(
            asn1.ASN1Value(asn1.TagClass.universal, False, asn1.TypeTagNumber.integer, b"\x01"),
            asn1.TagClass.universal,
            asn1.TypeTagNumber.octet_string,
        )


def test_extract_asn1_invalid_other_class():
    expected = (
        "Invalid ASN.1 tags, actual tag TagClass.application and number 1024, expecting class "
        "TagClass.application and number 512"
    )
    with pytest.raises(ValueError, match=re.escape(expected)):
        asn1.extract_asn1_tlv(
            asn1.ASN1Value(asn1.TagClass.application, True, 1024, b"\x01"), asn1.TagClass.application, 512
        )


def test_get_sequence_value():
    input_sequence = asn1.pack_asn1_sequence(
        [asn1.pack_asn1(asn1.TagClass.context_specific, True, 1, asn1.pack_asn1_integer(1))]
    )
    output_sequence = asn1.unpack_asn1_tagged_sequence(asn1.unpack_asn1(input_sequence)[0])

    actual = asn1.get_sequence_value(output_sequence, 1, "Structure", "field-name", asn1.unpack_asn1_integer)
    assert actual == 1


def test_get_sequence_value_raw():
    input_sequence = asn1.pack_asn1_sequence(
        [asn1.pack_asn1(asn1.TagClass.context_specific, True, 1, asn1.pack_asn1_integer(1))]
    )
    output_sequence = asn1.unpack_asn1_tagged_sequence(asn1.unpack_asn1(input_sequence)[0])

    actual = asn1.get_sequence_value(output_sequence, 1, "Structure")
    assert actual == asn1.ASN1Value(asn1.TagClass.universal, False, asn1.TypeTagNumber.integer, b"\x01")


def test_get_sequence_value_no_tag():
    input_sequence = asn1.pack_asn1_sequence(
        [asn1.pack_asn1(asn1.TagClass.context_specific, True, 1, asn1.pack_asn1_integer(1))]
    )
    output_sequence = asn1.unpack_asn1_tagged_sequence(asn1.unpack_asn1(input_sequence)[0])

    actual = asn1.get_sequence_value(output_sequence, 2, "Structure")
    assert actual is None


def test_get_sequence_value_failure_with_field_name():
    input_sequence = asn1.pack_asn1_sequence(
        [asn1.pack_asn1(asn1.TagClass.context_specific, True, 1, asn1.pack_asn1_integer(1))]
    )
    output_sequence = asn1.unpack_asn1_tagged_sequence(asn1.unpack_asn1(input_sequence)[0])

    expected = (
        "Failed unpacking field-name in Structure: Invalid ASN.1 OCTET STRING tags, actual tag class "
        "TagClass.universal and tag number TypeTagNumber.integer"
    )
    with pytest.raises(ValueError, match=re.escape(expected)):
        asn1.get_sequence_value(output_sequence, 1, "Structure", "field-name", asn1.unpack_asn1_octet_string)


def test_get_sequence_value_failure_without_field_name():
    input_sequence = asn1.pack_asn1_sequence(
        [asn1.pack_asn1(asn1.TagClass.context_specific, True, 1, asn1.pack_asn1_integer(1))]
    )
    output_sequence = asn1.unpack_asn1_tagged_sequence(asn1.unpack_asn1(input_sequence)[0])

    expected = (
        "Failed unpacking Structure: Invalid ASN.1 OCTET STRING tags, actual tag class "
        "TagClass.universal and tag number TypeTagNumber.integer"
    )
    with pytest.raises(ValueError, match=re.escape(expected)):
        asn1.get_sequence_value(output_sequence, 1, "Structure", unpack_func=asn1.unpack_asn1_octet_string)


@pytest.mark.parametrize("tag_class, constructed, tag_number, data, expected", ASN1_TAG_TESTS)
def test_pack_asn1_tlv(tag_class, constructed, tag_number, data, expected):
    actual = asn1.pack_asn1(tag_class, constructed, tag_number, data)
    assert actual == expected


@pytest.mark.parametrize("tag_class, constructed, tag_number, data, value", ASN1_TAG_TESTS)
def test_unpack_asn1_tlv(tag_class, constructed, tag_number, data, value):
    actual = asn1.unpack_asn1(value)

    assert actual[0].tag_class == tag_class
    assert actual[0].constructed == constructed
    assert actual[0].tag_number == tag_number
    assert actual[0].b_data == data
    assert actual[1] == b""


def test_unpack_asn1_tlv_remaining_data():
    actual = asn1.unpack_asn1(b"\x04\x01\x00\x01")

    assert actual[0].tag_class == asn1.TagClass.universal
    assert not actual[0].constructed
    assert actual[0].tag_number == asn1.TypeTagNumber.octet_string
    assert actual[0].b_data == b"\x00"
    assert actual[1] == b"\x01"


def test_pack_asn1_tlv_invalid_class():
    expected = "tag_class must be between 0 and 3"
    with pytest.raises(ValueError, match=re.escape(expected)):
        asn1.pack_asn1(10, False, 0, b"")  # type: ignore[arg-type]


def test_pack_asn1_bit_string():
    actual = asn1.pack_asn1_bit_string(b"\x01\x01")
    assert actual == b"\x03\x03\x00\x01\x01"


@pytest.mark.parametrize(
    "value, expected",
    [
        (b"\x03\x03\x00\x01\x01", b"\x01\x01"),
        (b"\x03\x03\x01\x01\x01", b"\x01\x00"),
    ],
)
def test_unpack_asn1_bit_string(value, expected):
    actual = asn1.unpack_asn1_bit_string(value[2:])
    assert actual == expected

    actual = asn1.unpack_asn1_bit_string(asn1.unpack_asn1(value)[0])
    assert actual == expected


def test_pack_asn1_general_string():
    actual = asn1.pack_asn1_general_string("cafe")
    assert actual == b"\x1B\x04\x63\x61\x66\x65"


def test_pack_asn1_enumerated():
    actual = asn1.pack_asn1_enumerated(1024)
    assert actual == b"\x0A\x02\x04\x00"


def test_unpack_asn1_enumerated():
    value = b"\x0A\x02\x04\x00"
    expected = 1024

    actual = asn1.unpack_asn1_enumerated(value[2:])
    assert actual == expected

    actual = asn1.unpack_asn1_enumerated(asn1.unpack_asn1(value)[0])
    assert actual == expected


def test_pack_asn1_general_string_encoding():
    actual = asn1.pack_asn1_general_string("café", encoding="utf-8")
    assert actual == b"\x1B\x05\x63\x61\x66\xC3\xA9"


def test_unpack_asn1_general_string():
    value = b"\x1B\x05\x63\x61\x66\xC3\xA9"
    expected = b"\x63\x61\x66\xC3\xA9"

    actual = asn1.unpack_asn1_general_string(value[2:])
    assert actual == expected

    actual = asn1.unpack_asn1_general_string(asn1.unpack_asn1(value)[0])
    assert actual == expected


@pytest.mark.parametrize("value, expected", INTEGER_TESTS)
def test_pack_asn1_integer(value, expected):
    actual = asn1.pack_asn1_integer(value)
    assert actual == expected

    actual_untagged = asn1.pack_asn1_integer(value, tag=False)
    assert actual_untagged == expected[2:]


@pytest.mark.parametrize("expected, value", INTEGER_TESTS)
def test_unpack_asn1_integer(expected, value):
    asn1_value = asn1.unpack_asn1(value)[0]
    actual = asn1.unpack_asn1_integer(asn1_value)
    assert actual == expected

    actual = asn1.unpack_asn1_integer(value[2:])
    assert actual == expected


@pytest.mark.parametrize("value, expected", OID_TESTS)
def test_pack_object_identifier(value, expected):
    actual = asn1.pack_asn1_object_identifier(value)
    assert actual == expected


def test_pack_object_identifier_invalid_value():
    expected = "An OID must have 2 or more elements split by '.'"
    with pytest.raises(ValueError, match=re.escape(expected)):
        asn1.pack_asn1_object_identifier("1")


@pytest.mark.parametrize("expected, value", OID_TESTS)
def test_unpack_object_identifier(expected, value):
    asn1_value = asn1.unpack_asn1(value)[0]
    actual = asn1.unpack_asn1_object_identifier(asn1_value)
    assert actual == expected

    actual = asn1.unpack_asn1_object_identifier(value[2:])
    assert actual == expected


def test_pack_asn1_sequence():
    inner_val = asn1.pack_asn1_octet_string(b"\x01\x02")
    actual = asn1.pack_asn1_sequence([inner_val, inner_val])

    assert actual == b"\x30\x08\x04\x02\x01\x02\x04\x02\x01\x02"


def test_unpack_asn1_sequence():
    value = b"\x30\x08\x04\x02\x01\x02\x04\x02\x01\x02"

    def actual_test(expected: typing.List) -> bool:
        assert isinstance(expected, list)

        for a in expected:
            assert isinstance(a, asn1.ASN1Value)
            assert a.tag_class == asn1.TagClass.universal
            assert not a.constructed
            assert a.tag_number == asn1.TypeTagNumber.octet_string
            assert a.b_data == b"\x01\x02"

        return True

    actual = asn1.unpack_asn1_sequence(value[2:])
    assert actual_test(actual)

    actual = asn1.unpack_asn1_sequence(asn1.unpack_asn1(value)[0])
    assert actual_test(actual)


def test_pack_asn1_octet_string():
    actual = asn1.pack_asn1_octet_string(b"\x01\x02")
    assert actual == b"\x04\x02\x01\x02"


def test_unpack_asn1_octet_string():
    value = b"\x04\x02\x01\x02"
    expected = b"\x01\x02"

    actual = asn1.unpack_asn1_octet_string(value[2:])
    assert actual == expected

    actual = asn1.unpack_asn1_octet_string(asn1.unpack_asn1(value)[0])
    assert actual == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (b"\x00", False),
        (b"\x01", True),
        (b"\x02", True),
    ],
)
def test_unpack_asn1_boolean(value, expected):
    actual = asn1.unpack_asn1_boolean(value)
    assert actual == expected

    actual = asn1.unpack_asn1_boolean(asn1.unpack_asn1(b"\x01\x01" + value)[0])
    assert actual == expected


@pytest.mark.parametrize(
    "value",
    [
        b"\x31\x39\x37\x30\x30\x31\x30\x31\x30\x30\x30\x30\x30\x30\x5A",
        b"\x31\x39\x37\x30\x30\x31\x30\x31\x30\x30\x30\x30\x30\x30\x2E\x30\x30\x30\x30\x30\x30\x5A",
    ],
)
def test_unpack_asn1_generalized_time(value):
    expected = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)

    actual = asn1.unpack_asn1_generalized_time(value)
    assert actual == expected

    packed_value = asn1.pack_asn1(asn1.TagClass.universal, False, asn1.TypeTagNumber.generalized_time, value)
    actual = asn1.unpack_asn1_generalized_time(asn1.unpack_asn1(packed_value)[0])
    assert actual == expected


def test_unpack_asn1_generalized_time_format_err():
    expected = "time data '1970' does not match format"
    with pytest.raises(ValueError, match=expected):
        asn1.unpack_asn1_generalized_time(b"\x31\x39\x37\x30")



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import json
import logging
import logging.config
import os

from spnego._context import (
    ContextProxy,
    ContextReq,
    IOVUnwrapResult,
    IOVWrapResult,
    UnwrapResult,
    WinRMWrapResult,
    WrapResult,
)
from spnego._credential import (
    Credential,
    CredentialCache,
    KerberosCCache,
    KerberosKeytab,
    NTLMHash,
    Password,
)
from spnego.auth import client, server
from spnego.exceptions import NegotiateOptions

__all__ = [
    "ContextProxy",
    "ContextReq",
    "Credential",
    "CredentialCache",
    "IOVUnwrapResult",
    "IOVWrapResult",
    "KerberosCCache",
    "KerberosKeytab",
    "NTLMHash",
    "Password",
    "NegotiateOptions",
    "UnwrapResult",
    "WinRMWrapResult",
    "WrapResult",
    "client",
    "server",
]


def _setup_logging(logger: logging.Logger) -> None:
    log_path = os.environ.get("PYSPNEGO_LOG_CFG", None)

    if log_path is not None and os.path.exists(log_path):  # pragma: no cover
        # log log config from JSON file
        with open(log_path, "rt") as f:
            config = json.load(f)

        logging.config.dictConfig(config)
    else:
        # no logging was provided
        logger.addHandler(logging.NullHandler())


logger = logging.getLogger(__name__)
_setup_logging(logger)



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pytest

import spnego._context as context


@pytest.mark.parametrize(
    "value, expected_domain, expected_user",
    [
        (None, None, None),
        ("username", None, "username"),
        ("domain\\username", "domain", "username"),
        ("domain\\username\\part", "domain", "username\\part"),
        ("username@domain", None, "username@domain"),
    ],
)
def test_split_username(value, expected_domain, expected_user):
    actual_domain, actual_user = context.split_username(value)

    assert actual_domain == expected_domain
    assert actual_user == expected_user


@pytest.mark.parametrize(
    "value, expected",
    [
        (context.GSSMech.ntlm, "ntlm"),
        (context.GSSMech.spnego, "spnego"),
        (context.GSSMech.kerberos, "kerberos"),
        (context.GSSMech._ms_kerberos, "kerberos"),
        (context.GSSMech._kerberos_draft, "kerberos"),
        (context.GSSMech._iakerb, "kerberos"),
    ],
)
def test_gss_mech_common_name(value, expected):
    actual = context.GSSMech(value).common_name

    assert actual == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (context.GSSMech.ntlm, False),
        (context.GSSMech.spnego, False),
        (context.GSSMech.kerberos, True),
        (context.GSSMech._ms_kerberos, True),
        (context.GSSMech._kerberos_draft, True),
        (context.GSSMech._iakerb, True),
    ],
)
def test_gss_mech_is_kerberos_oid(value, expected):
    assert context.GSSMech(value).is_kerberos_oid == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        ("1.3.6.1.4.1.311.2.2.10", context.GSSMech.ntlm),
        ("1.3.6.1.5.5.2", context.GSSMech.spnego),
        ("1.2.840.113554.1.2.2", context.GSSMech.kerberos),
        ("1.2.840.48018.1.2.2", context.GSSMech._ms_kerberos),
    ],
)
def test_gss_mech_from_oid(value, expected):
    actual = context.GSSMech.from_oid(value)

    assert actual == expected


def test_gss_mech_from_oid_invalid():
    with pytest.raises(ValueError, match="is not a valid GSSMech OID"):
        context.GSSMech.from_oid("1.2.3.4.5")



# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pytest

import spnego._spnego as sp
from spnego._asn1 import TagClass, pack_asn1, pack_asn1_object_identifier
from spnego._context import GSSMech
from spnego._kerberos import KrbApRep, KrbApReq
from spnego._ntlm_raw.messages import Challenge, MessageType

from .conftest import get_data


@pytest.mark.parametrize(
    "value, expected",
    [
        (GSSMech.kerberos.value, b"\x30\x0B\x06\x09\x2A\x86\x48\x86\xF7\x12\x01\x02\x02"),
        (
            [GSSMech.kerberos.value, GSSMech.ntlm.value],
            b"\x30\x17\x06\x09\x2A\x86\x48\x86"
            b"\xF7\x12\x01\x02\x02\x06\x0A\x2B"
            b"\x06\x01\x04\x01\x82\x37\x02\x02"
            b"\x0A",
        ),
    ],
)
def test_pack_mech_type_list(value, expected):
    actual = sp.pack_mech_type_list(value)
    import base64

    print(base64.b16encode(actual).decode())
    assert actual == expected


def test_spnego_context_flags_native_labels():
    actual = sp.ContextFlags.native_labels()

    assert isinstance(actual, dict)
    assert actual[sp.ContextFlags.deleg] == "delegFlag"


def test_spnego_neg_state_native_labels():
    actual = sp.NegState.native_labels()

    assert isinstance(actual, dict)
    assert actual[sp.NegState.accept_incomplete] == "accept-incomplete"


def test_initial_context_token_unknown_token():
    token = sp.InitialContextToken(GSSMech.ntlm, b"\x00\x00\x00\x00")

    assert isinstance(token, sp.InitialContextToken)
    assert token.this_mech == GSSMech.ntlm.value
    assert token.inner_context_token == b"\x00\x00\x00\x00"
    assert token.token == b"\x00\x00\x00\x00"


def test_initial_context_token_unknown_mech():
    token = sp.InitialContextToken("1.2.3.4.5", b"\x00\x00\x00\x00")

    assert isinstance(token, sp.InitialContextToken)
    assert token.this_mech == "1.2.3.4.5"
    assert token.inner_context_token == b"\x00\x00\x00\x00"
    assert token.token == b"\x00\x00\x00\x00"


def test_unpack_initial_context_token_unknown_mech():
    data = pack_asn1(TagClass.application, True, 0, pack_asn1_object_identifier("1.2.3.4.5") + b"\x00\x00\x00\x00")
    token = sp.unpack_token(data)

    assert token == data


def test_unpack_initial_context_token_invalid_application_tag():
    data = pack_asn1(TagClass.application, True, 1, b"\x00\x00\x00\x00")

    expected = "Expecting a tag number of 0 not 1 for InitialContextToken"
    with pytest.raises(ValueError, match=expected):
        sp.unpack_token(data)


def test_unpack_initial_context_token_invalid_context_specific_tag():
    data = pack_asn1(TagClass.context_specific, True, 2, b"\x00\x00\x00\x00")

    expected = "Unknown NegotiationToken CHOICE 2, only expecting 0 or 1"
    with pytest.raises(ValueError, match=expected):
        sp.unpack_token(data)


def test_unpack_neg_token_init():
    data = get_data("initial_context_token_neg_token_init")
    actual = sp.unpack_token(data)

    assert isinstance(actual, sp.NegTokenInit)
    assert actual.hint_address is None
    assert actual.hint_name is None
    assert actual.mech_list_mic is None
    assert isinstance(actual.mech_token, bytes)
    assert actual.mech_types == ["1.2.840.113554.1.2.2", "1.3.6.1.4.1.311.2.2.10"]
    assert actual.req_flags is None

    actual = sp.unpack_token(data, unwrap=True)
    assert isinstance(actual, sp.InitialContextToken)
    assert actual.this_mech == GSSMech.spnego.value
    assert isinstance(actual.token, sp.NegTokenInit)


def test_unpack_neg_token_init2():
    data = get_data("initial_context_token_neg_token_init2")
    actual = sp.unpack_token(data)

    assert isinstance(actual, sp.NegTokenInit)
    assert actual.hint_address is None
    assert actual.hint_name == b"not_defined_in_RFC4178@please_ignore"
    assert actual.mech_list_mic is None
    assert actual.mech_token is None
    assert actual.mech_types == [
        "1.3.6.1.4.1.311.2.2.30",
        "1.2.840.48018.1.2.2",
        "1.2.840.113554.1.2.2",
        "1.2.840.113554.1.2.2.3",
        "1.3.6.1.4.1.311.2.2.10",
    ]
    assert actual.req_flags is None

    actual = sp.unpack_token(data, unwrap=True)
    assert isinstance(actual, sp.InitialContextToken)
    assert actual.this_mech == GSSMech.spnego.value
    assert isinstance(actual.token, sp.NegTokenInit)


def test_unpack_neg_token_resp():
    data = get_data("neg_token_resp")
    actual = sp.unpack_token(data)

    assert isinstance(actual, sp.NegTokenResp)
    assert actual.mech_list_mic is None
    assert actual.neg_state == sp.NegState.accept_complete
    assert isinstance(actual.response_token, bytes)
    assert actual.supported_mech == GSSMech.kerberos.value

    actual = sp.unpack_token(data, unwrap=True)
    assert isinstance(actual, sp.NegTokenResp)


def test_unpack_krb_ap_req():
    data = get_data("initial_context_token_krb_ap_req")
    actual = sp.unpack_token(data)

    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)
    assert isinstance(actual, sp.InitialContextToken)
    assert actual.this_mech == GSSMech.kerberos.value
    assert isinstance(actual.token, KrbApReq)


def test_unpack_krb_ap_rep():
    data = get_data("initial_context_token_krb_ap_rep")
    actual = sp.unpack_token(data)

    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)
    assert isinstance(actual, sp.InitialContextToken)
    assert actual.this_mech == GSSMech.kerberos.value
    assert isinstance(actual.token, KrbApRep)


def test_unpack_ntlm():
    data = get_data("ntlm_challenge")
    actual = sp.unpack_token(data)

    assert actual == data

    actual = sp.unpack_token(data, unwrap=True)
    assert isinstance(actual, Challenge)
    assert actual.MESSAGE_TYPE == MessageType.challenge


def test_pack_neg_token_init():
    token = sp.NegTokenInit(
        [GSSMech.kerberos.value, GSSMech.ntlm.value],
        sp.ContextFlags.anon,
        b"\x00\x00\x00\x00",
        mech_list_mic=b"\x01\x01\x01\x01",
    )

    actual = token.pack()
    assert (
        actual == b"\x60\x3D\x06\x06\x2B\x06\x01\x05\x05\x02\xA0\x33\x30\x31\xA0\x19"
        b"\x30\x17\x06\x09\x2A\x86\x48\x86\xF7\x12\x01\x02\x02\x06\x0A\x2B"
        b"\x06\x01\x04\x01\x82\x37\x02\x02\x0A\xA1\x04\x03\x02\x00\x04\xA2"
        b"\x06\x04\x04\x00\x00\x00\x00\xA3\x06\x04\x04\x01\x01\x01\x01"
    )

    token = sp.unpack_token(actual)
    assert token.mech_types == ["1.2.840.113554.1.2.2", "1.3.6.1.4.1.311.2.2.10"]
    assert token.req_flags == sp.ContextFlags.deleg
    assert token.mech_token == b"\x00\x00\x00\x00"
    assert token.hint_name is None
    assert token.hint_address is None
    assert token.mech_list_mic == b"\x01\x01\x01\x01"


def test_pack_neg_token_init2():
    token = sp.NegTokenInit(
        [GSSMech.kerberos.value, GSSMech.ntlm.value],
        sp.ContextFlags.anon,
        mech_token=b"\x00\x00\x00\x00",
        hint_name=b"spn",
        hint_address=b"not_defined_in_RFC4178@please_ignore",
        mech_list_mic=b"\x01\x01\x01\x01",
    )

    actual = token.pack()

    assert (
        actual == b"\x60\x70\x06\x06\x2B\x06\x01\x05\x05\x02\xA0\x66\x30\x64\xA0\x19"
        b"\x30\x17\x06\x09\x2A\x86\x48\x86\xF7\x12\x01\x02\x02\x06\x0A\x2B"
        b"\x06\x01\x04\x01\x82\x37\x02\x02\x0A\xA1\x04\x03\x02\x00\x04\xA2"
        b"\x06\x04\x04\x00\x00\x00\x00\xA3\x31\x30\x2F\xA0\x05\x1B\x03\x73"
        b"\x70\x6E\xA1\x26\x04\x24\x6E\x6F\x74\x5F\x64\x65\x66\x69\x6E\x65"
        b"\x64\x5F\x69\x6E\x5F\x52\x46\x43\x34\x31\x37\x38\x40\x70\x6C\x65"
        b"\x61\x73\x65\x5F\x69\x67\x6E\x6F\x72\x65\xA4\x06\x04\x04\x01\x01"
        b"\x01\x01"
    )

    token = sp.unpack_token(actual)
    assert token.mech_types == ["1.2.840.113554.1.2.2", "1.3.6.1.4.1.311.2.2.10"]
    assert token.req_flags == sp.ContextFlags.deleg
    assert token.mech_token == b"\x00\x00\x00\x00"
    assert token.hint_name == b"spn"
    assert token.hint_address == b"not_defined_in_RFC4178@please_ignore"
    assert token.mech_list_mic == b"\x01\x01\x01\x01"


def test_pack_neg_token_resp():
    token = sp.NegTokenResp(sp.NegState.request_mic, GSSMech.ntlm.value, b"\x00\x00\x00\x00", b"\x01\x01\x01\x01")

    actual = token.pack()
    assert (
        actual == b"\xA1\x25\x30\x23\xA0\x03\x0A\x01\x03\xA1\x0C\x06\x0A\x2B\x06\x01"
        b"\x04\x01\x82\x37\x02\x02\x0A\xA2\x06\x04\x04\x00\x00\x00\x00\xA3"
        b"\x06\x04\x04\x01\x01\x01\x01"
    )

    token = sp.unpack_token(actual)
    assert token.neg_state == sp.NegState.request_mic
    assert token.supported_mech == GSSMech.ntlm.value
    assert token.response_token == b"\x00\x00\x00\x00"
    assert token.mech_list_mic == b"\x01\x01\x01\x01"



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import sys

import spnego.channel_bindings as cb

TEST_DATA = cb.GssChannelBindings(
    cb.AddressType.inet, b"\x01\x02\x03\x04", cb.AddressType.unspecified, b"\x05\x06\x07\x08", b"caf\xC3\xA9"
)

TEST_B_DATA = (
    b"\x02\x00\x00\x00\x04\x00\x00\x00\x01\x02\x03\x04"
    b"\x00\x00\x00\x00\x04\x00\x00\x00\x05\x06\x07\x08"
    b"\x05\x00\x00\x00caf\xC3\xA9"
)


def test_channel_bindings_pack():
    actual = TEST_DATA.pack()

    assert actual == TEST_B_DATA


def test_channel_bindings_none_pack():
    actual = cb.GssChannelBindings().pack()

    assert actual == b"\x00\x00\x00\x00\x00\x00\x00\x00" b"\x00\x00\x00\x00\x00\x00\x00\x00" b"\x00\x00\x00\x00"


def test_channel_bindings_unpack():
    actual = cb.GssChannelBindings.unpack(TEST_B_DATA)

    assert isinstance(actual, cb.GssChannelBindings)

    assert actual.initiator_addrtype == cb.AddressType.inet
    assert actual.initiator_address == b"\x01\x02\x03\x04"
    assert actual.acceptor_addrtype == cb.AddressType.unspecified
    assert actual.acceptor_address == b"\x05\x06\x07\x08"
    assert actual.application_data == b"caf\xC3\xA9"


def test_channel_bindings_str():
    actual = str(TEST_DATA)

    assert (
        actual == r"GssChannelBindings initiator_addr(AddressType.inet|b'\x01\x02\x03\x04') | "
        r"acceptor_addr(AddressType.unspecified|b'\x05\x06\x07\x08') | "
        r"application_data(b'caf\xc3\xa9')"
    )


def test_channel_bindings_repr():
    actual = repr(TEST_DATA)

    assert (
        actual == r"spnego.channel_bindings.GssChannelBindings initiator_addrtype=<AddressType.inet: 2>|"
        r"initiator_address=b'\x01\x02\x03\x04'|"
        r"acceptor_addrtype=<AddressType.unspecified: 0>|"
        r"acceptor_address=b'\x05\x06\x07\x08'|"
        r"application_data=b'caf\xc3\xa9'"
    )


def test_channel_bindings_eq():
    assert TEST_DATA == cb.GssChannelBindings.unpack(TEST_B_DATA)

    other = cb.GssChannelBindings.unpack(TEST_B_DATA)
    assert TEST_DATA == other

    assert TEST_DATA != 1

    other.acceptor_address = b"new"
    assert TEST_DATA != other
